from typing import Union, Optional, TYPE_CHECKING
from metaflow import current

from .identity_utils import (
    TaskIdentifier,
    FlowNotRunningException,
)
from collections import namedtuple


if TYPE_CHECKING:
    import metaflow


ResolvedTask = namedtuple(
    "ResolvedTask",
    [
        "flow_name",
        "run_id",
        "step_name",
        "task_id",
        "is_foreach",
        "current_attempt",
        "pathspec",
        "control_task_pathspec",
    ],
)

OriginInfo = namedtuple(
    "OriginInfo", ["origin_run_id", "origin_task_id", "origin_attempt", "is_foreach"]
)


def resolve_pathspec_for_flowspec(
    run: "metaflow.FlowSpec" = None,
):
    if not current.is_running_flow:
        raise FlowNotRunningException
    control_task_pathspec = None
    if getattr(current, "parallel", None):
        if current.parallel.control_task_id:
            control_task_pathspec = "/".join(
                [
                    current.flow_name,
                    current.run_id,
                    current.step_name,
                    current.parallel.control_task_id,
                ]
            )
    return ResolvedTask(
        current.flow_name,
        current.run_id,
        current.step_name,
        current.task_id,
        run.index is not None,
        current.retry_count,
        current.pathspec,
        control_task_pathspec,
    )


def resolve_storage_backend(run: "metaflow.FlowSpec" = None):
    if not current.is_running_flow:
        raise FlowNotRunningException
    return run._datastore._storage_impl


def resolve_task_identifier(
    run: "metaflow.FlowSpec",
    gang_scheduled_task=False,
    gang_schedule_task_idf_index=0,
):
    if not current.is_running_flow:
        raise FlowNotRunningException

    if gang_scheduled_task:
        return TaskIdentifier.for_parallel_task_index(run, gang_schedule_task_idf_index)
    return TaskIdentifier.from_flowspec(run)



# from functools import wraps
from datetime import datetime
import os

from .exceptions import CheckpointException
from ..utils import flowspec_utils
from ..card_utils import CardDecoratorInjector, AsyncPeriodicRefresher
from .cards.checkpoint_lister import CheckpointListRefresher, CheckpointsCollector
from .cards import create_checkpoint_card, null_card
from .lineage import checkpoint_load_related_metadata, trace_lineage
from .constructors import (
    DEFAULT_NAME,
    load_checkpoint,
    _instantiate_checkpoint,
)
from .core import (
    ScopeResolver,
    CheckpointLoadPolicy,
)

# from .lineage import checkpoint_load_related_metadata, trace_lineage
from ..datastructures import CheckpointArtifact
from .constants import (
    CHECKPOINT_TAG_PREFIX,
    CHECKPOINT_TASK_IDENTIFIER_ENV_VAR_NAME,
    CHECKPOINT_UID_ENV_VAR_NAME,
    TASK_CHECKPOINTS_ARTIFACT_NAME,
    TASK_LATEST_CHECKPOINT_ARTIFACT_NAME,
    DEFAULT_STORAGE_FORMAT,
)

# from .cards import CardDecoratorInjector, create_checkpoint_card, null_card
from metaflow.decorators import StepDecorator, FlowDecorator
from metaflow.flowspec import INTERNAL_ARTIFACTS_SET
from .final_api import Checkpoint
from typing import List, Dict, Union, Tuple, Optional, Callable, TYPE_CHECKING
from functools import wraps, partial
import tempfile
import json

if TYPE_CHECKING:
    import metaflow


try:
    unicode
except NameError:
    unicode = str
    basestring = str


def _store_checkpoint_ref_as_data_artifact(
    flow,
    attempt,
    checkpointer: "CurrentCheckpointer",
):
    _chckpts = []
    # There are two code paths we will try out:
    # One code path will try and get the latest checkpoint since
    # Another one will try and retrive all of them. On reason we do this is because
    # `list` call to the datastore can return things out of order and since we are also filtering for a
    # max value in the list, we might end up with a checkpoint that is not the latest.

    # This GLITCH is that if a user, is running Flows concurrently and they are all updating the value of
    # `latest` then it can surely create a problem.
    latest_chckpt = (
        checkpointer._default_checkpointer._checkpointer._get_latest_checkpoint(
            safe=True
        )
    )

    for checkpoint in checkpointer._default_checkpointer.list(
        attempt=attempt, as_dict=False, within_task=True
    ):
        _chckpts.append(checkpoint)

    time_sorted = lambda z: sorted(
        z, key=lambda x: datetime.fromisoformat(x.created_on), reverse=True
    )
    _chckpts = time_sorted(_chckpts)
    # TASK_CHECKPOINTS_ARTIFACT_NAME is a list of dictionaries that will be stored in the datastore.
    # This is a variable hidden from the user but will be used to locate the checkpoints
    # in the datastore.
    setattr(
        flow,
        TASK_CHECKPOINTS_ARTIFACT_NAME,
        [c.to_dict() for c in _chckpts],
    )
    setattr(
        flow,
        TASK_LATEST_CHECKPOINT_ARTIFACT_NAME,
        None,
    )
    if latest_chckpt:
        setattr(
            flow,
            TASK_LATEST_CHECKPOINT_ARTIFACT_NAME,
            latest_chckpt.to_dict(),
        )


def warning_message(message, logger=None, ts=False, prefix="[@checkpoint]"):
    msg = "%s %s" % (prefix, message)
    if logger:
        logger(msg, timestamp=ts, bad=True)


class CurrentCheckpointer:
    @property
    def task_identifier(self):
        return self._task_identifier

    @property
    def directory(self):
        return self._temp_chckpt_dir.name

    @property
    def is_loaded(self):
        return self._loaded_checkpoint is not None

    @property
    def info(self):
        return self._loaded_checkpoint

    def __init__(
        self,
        flow,
        task_identifier,
        resolved_scope,
        logger,
        gang_scheduled_task=False,
        temp_dir_root=None,
    ) -> None:
        from metaflow import current

        self._temp_dir_root = temp_dir_root
        self._resolved_scope = resolved_scope
        self._logger = logger
        self._loaded_checkpoint = None
        # Ensure that if a tempdir root path is provided and nothing
        # exists then we end up creating that path. This helps ensure
        # that rouge paths with arbirary Filesystems get created before
        # temp dirs exists.
        if temp_dir_root is not None:
            if not os.path.exists(temp_dir_root):
                os.makedirs(temp_dir_root, exist_ok=True)
        self._temp_chckpt_dir = tempfile.TemporaryDirectory(
            prefix="metaflow_checkpoint_", dir=self._temp_dir_root
        )
        self._task_identifier = task_identifier
        self._default_checkpointer = _instantiate_checkpoint(
            Checkpoint(),
            flow=flow,
            task_identifier=task_identifier,
            scope=resolved_scope,
            gang_scheduled_task=gang_scheduled_task,
        )
        os.environ[CHECKPOINT_TASK_IDENTIFIER_ENV_VAR_NAME] = self._task_identifier
        os.environ[CHECKPOINT_UID_ENV_VAR_NAME] = str(
            self._default_checkpointer._checkpointer._checkpoint_uid
        )

    def _setup_task_first_load(self, load_policy, flow):
        checkpoint = None
        if load_policy == "eager":
            checkpoint = CheckpointLoadPolicy.eager(
                self._default_checkpointer._checkpointer._checkpoint_datastore,
                flow,
            )
        elif load_policy == "fresh":
            checkpoint = CheckpointLoadPolicy.fresh(
                self._default_checkpointer._checkpointer._checkpoint_datastore,
                flow,
            )
        if checkpoint is None:
            return None

        warning_message(
            "Loading the following checkpoint:\n\t[pathspec] %s\n\t[key] %s\n\t[created on] %s\n\t[url] %s"
            % (
                checkpoint.pathspec,
                checkpoint.key,
                checkpoint.created_on,
                checkpoint.url,
            ),
            logger=self._logger,
            ts=False,
        )
        load_checkpoint(checkpoint=checkpoint, local_path=self.directory)
        self._loaded_checkpoint = checkpoint
        return checkpoint

    def save(
        self,
        path: Optional[Union[str, os.PathLike]] = None,
        name: Optional[str] = DEFAULT_NAME,
        metadata: Optional[Dict] = {},
        latest: bool = True,
        storage_format: str = DEFAULT_STORAGE_FORMAT,
    ):
        """
        Saves the checkpoint to the datastore.

        Parameters
        ----------
        path : Optional[Union[str, os.PathLike]], default: None
            The path to save the checkpoint. Accepts a file path or a directory path.
                - If a directory path is provided, all the contents within that directory will be saved.
                When a checkpoint is reloaded during task retries, `the current.checkpoint.directory` will
                contain the contents of this directory.
                - If a file path is provided, the file will be directly saved to the datastore (with the same filename).
                When the checkpoint is reloaded during task retries, the file with the same name will be available in the
                `current.checkpoint.directory`.
                - If no path is provided then the `current.checkpoint.directory` will be saved as the checkpoint.

        name : Optional[str], default: "mfchckpt"
            The name of the checkpoint.

        metadata : Optional[Dict], default: {}
            Any metadata that needs to be saved with the checkpoint.

        latest : bool, default: True
            If True, the checkpoint will be marked as the latest checkpoint.
            This helps determine if the checkpoint gets loaded when the task restarts.

        storage_format : str, default: files
            If `tar`, the contents of the directory will be tarred before saving to the datastore.
            If `files`, saves directory directly to the datastore.

        """
        if path is None:
            path = self.directory
        return self._default_checkpointer.save(
            path=path,
            name=name,
            metadata=metadata,
            latest=latest,
            storage_format=storage_format,
        )

    def list(
        self,
        name: Optional[str] = None,
        task: Optional[Union[str, "metaflow.Task"]] = None,
        attempt: Optional[int] = None,
        within_task: bool = True
        # `within_task` is a hidden option. If set to False, it will list all checkpoints in "scope" (current namespace).
        # If true, it will list all checkpoints created within the task
    ):
        return self._default_checkpointer.list(
            name=name,
            task=task,
            attempt=attempt,
            as_dict=True,
            within_task=within_task,
        )

    def cleanup(self):
        self._temp_chckpt_dir.cleanup()

    def refresh_directory(self):
        self.cleanup()
        self._temp_chckpt_dir = tempfile.TemporaryDirectory(
            prefix="metaflow_checkpoint_", dir=self._temp_dir_root
        )

    def load(
        self,
        reference: Union[str, Dict, CheckpointArtifact],
        path: Optional[str] = None,
    ):
        """
        loads a checkpoint reference from the datastore. (resembles a read op)

        This can have two meanings:
            - If the path is provided, it will load the checkpoint in the provided path
            - If no path is provided, it will load the checkpoint in the default directory

        Parameters
        ----------

        `reference` :
            - can be a string, dict or a CheckpointArtifact object:
                - string: a string reference to the checkpoint
                - dict: a dictionary form of the CheckpointArtifact
                - CheckpointArtifact: a CheckpointArtifact object reference to the checkpoint
        """
        if path is None:
            self.refresh_directory()
            path = self.directory
        return Checkpoint().load(reference, path=path)


def merge_dicts_with_precedence(*args: dict) -> dict:
    """
    Merges multiple dictionaries, respecting the order of precedence.

    This function takes any number of dictionary arguments and merges them into a single dictionary.
    If the same key exists in multiple dictionaries, the value from the dictionary that appears
    last in the argument list takes precedence, except where the value is None, in which case
    the search continues in the earlier dictionaries for a non-None value.

    The operation is not recursive and will only consider top-level keys.

    Parameters:
    - args: A variable number of dictionary arguments. Each argument must be a dictionary.

    Returns:
    - dict: A single dictionary that results from merging the input dictionaries according to their order of precedence.

    Examples:
    - merge_dicts_with_precedence(defaults, attrs)
      Here, `defaults` is a dictionary of default values, and `attrs` contains override values.
      Any None values in `attrs` will result in values from `defaults` being used.

    - merge_dicts_with_precedence(defaults, global_config, attrs)
      In this scenario, `global_config` can override `defaults`, and `attrs` can override both
      `defaults` and `global_config`. The order of arguments defines the precedence.

    Note:
    The function behaves differently if the order of the arguments changes, reflecting the
    precedence of the values set based on their position in the argument list.
    """
    unique_set_of_keys = set()
    for dictionary in args:
        unique_set_of_keys.update(dictionary.keys())

    final_dict = {}
    for key in unique_set_of_keys:
        for dictionary in args:
            if key in dictionary and dictionary[key] is not None:
                final_dict[key] = dictionary[key]
    return final_dict


def _greater_than_one_set(*args):
    return len([a for a in args if a]) > 1


class CheckpointDecorator(StepDecorator):
    """
    Enables checkpointing for a step.


    Parameters
    ----------
    load_policy : str, default: "fresh"
        The policy for loading the checkpoint. The following policies are supported:
            - "eager": Loads the the latest available checkpoint within the namespace.
            With this mode, the latest checkpoint written by any previous task (can be even a different run) of the step
            will be loaded at the start of the task.
            - "none": Do not load any checkpoint
            - "fresh": Loads the lastest checkpoint created within the running Task.
            This mode helps loading checkpoints across various retry attempts of the same task.
            With this mode, no checkpoint will be loaded at the start of a task but any checkpoints
            created within the task will be loaded when the task is retries execution on failure.

    temp_dir_root : str, default: None
        The root directory under which `current.checkpoint.directory` will be created.


    MF Add To Current
    -----------------
    checkpoint -> metaflow_extensions.obcheckpoint.plugins.machine_learning_utilities.checkpoints.decorator.CurrentCheckpointer
        The `@checkpoint` decorator makes saving/loading checkpoints available through the `current.checkpoint`.
        The object exposes `save`/`load`/`list` methods for saving/loading checkpoints.

        You can check if a checkpoint is loaded by `current.checkpoint.is_loaded` and get the checkpoint information
        by using `current.checkpoint.info`. The `current.checkpoint.directory` returns the path to the checkpoint directory
        where the checkpoint maybe loaded or saved.

        Usage (Saving Checkpoints):
        -------
        ```
        @checkpoint
        @step
        def train(self):
            model = create_model(self.parameters, checkpoint_path = None)
            for i in range(self.epochs):
                # some training logic
                loss = model.train(self.dataset)
                if i % 10 == 0:
                    model.save(
                        current.checkpoint.directory,
                    )
                    # saves the contents of the `current.checkpoint.directory` as a checkpoint
                    # and returns a reference dictionary to the checkpoint saved in the datastore
                    self.latest_checkpoint = current.checkpoint.save(
                        name="epoch_checkpoint",
                        metadata={
                            "epoch": i,
                            "loss": loss,
                        }
                    )
        ```
        Usage (Using Loaded Checkpoints):
        -------
        ```
        @retry(times=3)
        @checkpoint
        @step
        def train(self):
            # Assume that the task has restarted and the previous attempt of the task
            # saved a checkpoint
            checkpoint_path = None
            if current.checkpoint.is_loaded: # Check if a checkpoint is loaded
                print("Loaded checkpoint from the previous attempt")
                checkpoint_path = current.checkpoint.directory

            model = create_model(self.parameters, checkpoint_path = checkpoint_path)
            for i in range(self.epochs):
                ...
        ```

        @@ Returns
        -------
        CurrentCheckpointer
            The object for handling checkpointing within a step.
    """

    _task_identifier = None

    name = "checkpoint"

    defaults = {
        # `load_policy` defines the policy for the checkpoint loading during the execution of different runs.
        # It can be : ["eager", "none", "fresh"],
        "load_policy": "fresh",  #
        "temp_dir_root": None,  # Root directory for the temporary checkpoint directory.
    }

    LOAD_POLCIES = [
        # Check the `CheckpointLoadPolicy` for more documentation on these modes.
        "eager",
        "none",
        "fresh",
    ]

    def _resolve_settings(self):
        return merge_dicts_with_precedence(
            {"load_policy": "fresh"},
            self.attributes,
        )

    def step_init(
        self, flow, graph, step_name, decorators, environment, flow_datastore, logger
    ):
        self.deco_injector = CardDecoratorInjector()
        if (
            self.attributes["load_policy"] is not None
            and self.attributes["load_policy"] not in self.LOAD_POLCIES
        ):
            raise CheckpointException(
                "`load_policy` of %s is not supported. Supported policies are %s"
                % (self.attributes["load_policy"], ", ".join(self.LOAD_POLCIES))
            )

        # We add to INTERNAL_ARTIFACTS_SET here because the decorator adds internal artifacts to the
        # flow. Adding to INTERNAL_ARTIFACTS_SET avoids having any crashes when `merge_artifacts`
        # is called.
        INTERNAL_ARTIFACTS_SET.update(
            [TASK_CHECKPOINTS_ARTIFACT_NAME, TASK_LATEST_CHECKPOINT_ARTIFACT_NAME]
        )
        self._flow_datastore = flow_datastore
        self._logger = logger

        self.deco_injector.attach_card_decorator(
            flow,
            step_name,
            CheckpointListRefresher.CARD_ID,
            "blank",
            refresh_interval=2,
        )
        self._chkptr = None
        self._collector_thread = None

    def task_exception(
        self, exception, step_name, flow, graph, retry_count, max_user_code_retries
    ):
        if self._collector_thread is not None:
            self._collector_thread.stop()

        if self._chkptr is not None:
            _store_checkpoint_ref_as_data_artifact(flow, retry_count, self._chkptr)
            self._chkptr.cleanup()
            self._chkptr = None

    def task_pre_step(
        self,
        step_name,
        task_datastore,
        metadata,
        run_id,
        task_id,
        flow,
        graph,
        retry_count,
        max_user_code_retries,
        ubf_context,
        inputs,
    ):
        from metaflow import current

        settings = self._resolve_settings()
        load_policy = settings["load_policy"]
        self._load_policy = load_policy
        resolved_scope = self._resolve_scope()
        gang_scheduled_task = graph[step_name].parallel_step

        temp_dir_root = settings.get(
            "temp_dir_root",
        )

        if gang_scheduled_task and not getattr(
            current.parallel, "control_task_id", None
        ):
            # gang_scheduled_task's need a `current.parallel.control_task_id` to be set.
            # This is needed so that the metadata_store of the CheckpointDatastore is set to
            # the control task and hence when the gang restarts for any load policy, we are able
            # to load the checkpoint that was written by the control task.
            # Ideally in distributed data parallel scenarios, it is important to have a control task
            # that is writing the checkpoints and the other workers are reading from it when any restart
            # happens.
            gang_scheduled_task = False
            warning_message(
                "The task is a gang scheduled task but the control task id is not set. Metaflow version needs to be upgrade",
                logger=self._logger,
                ts=False,
            )

        default_task_identifier = flowspec_utils.resolve_task_identifier(
            run=flow,
            gang_scheduled_task=gang_scheduled_task,
            gang_schedule_task_idf_index=0,
            # TODO [POST RELEASE]: Make this a little more customizable in the future.
            # since the @parallel tasks can even be HPO style tasks instead
            # of gang scheduled tasks.
        )

        if gang_scheduled_task:
            # A step with an @parallel will mean that the decorator is follow gang scheduling semantics
            # so all workers and control task will be writing to the same path (i.e. the same
            # task identifier). If users wish to write checkpoints that are across different
            # workers, they should ensure that the `name` is differently set in the `save`
            # method so that checkpoints don't get overwritten.
            warning_message(
                (
                    "The step has a @parallel decorator and so checkpoints will be treated as if"
                    "they are coming from the same task. The checkpoints will be written/loaded from the control task."
                    "All tasks within this step will write to the same path i.e. the path of the control task."
                ),
                logger=self._logger,
                ts=False,
            )
        self._loaded_checkpoint = self._setup_checkpointer(
            flow,
            default_task_identifier,
            resolved_scope,
            load_policy,
            gang_scheduled_task=gang_scheduled_task,
            temp_dir_root=temp_dir_root,
        )
        self._loaded_checkpoint_lineage = []
        if self._loaded_checkpoint is not None:
            entries = checkpoint_load_related_metadata(
                self._loaded_checkpoint, retry_count
            )
            metadata.register_metadata(run_id, step_name, task_id, entries)
            checkpoint_list = trace_lineage(flow, self._loaded_checkpoint)
            if checkpoint_list:
                self._loaded_checkpoint_lineage = checkpoint_list
        self._setup_current()

    def task_decorate(
        self, step_func, flow, graph, retry_count, max_user_code_retries, ubf_context
    ):
        # Wrap the step_func in a function that will write to current.card["checkpoint_info"] the lineage card
        # and then call the step_func.
        self._collector_thread = CheckpointsCollector(
            CheckpointListRefresher(
                self._loaded_checkpoint,
                self._loaded_checkpoint_lineage,
                self._load_policy,
            ),
            interval=3,
        )

        def _wrapped_step_func(_collector_thread, *args, **kwargs):

            _collector_thread.start()
            try:
                return step_func(*args, **kwargs)
            finally:
                _collector_thread.stop()

        return partial(_wrapped_step_func, self._collector_thread)

    def task_post_step(
        self, step_name, flow, graph, retry_count, max_user_code_retries
    ):
        if self._collector_thread is not None:
            self._collector_thread.stop()

        if self._chkptr is not None:
            _store_checkpoint_ref_as_data_artifact(flow, retry_count, self._chkptr)
            self._chkptr.cleanup()
            self._chkptr = None

    def _resolve_scope(self):
        from metaflow import Run, current

        filtered_tags = [
            t
            for t in Run("%s/%s" % (current.flow_name, current.run_id)).tags
            if CHECKPOINT_TAG_PREFIX in t
        ]
        if len(filtered_tags) > 0:
            warning_message(
                "The Run has tags set with the special '%s' prefix; The checkpoints for this task will be namespaced under %s"
                % (CHECKPOINT_TAG_PREFIX, filtered_tags[0]),
                logger=self._logger,
                ts=False,
            )
            return ScopeResolver.from_tags(filtered_tags)
        else:
            return ScopeResolver.from_namespace()

    def _setup_checkpointer(
        self,
        flow,
        default_task_identifier,
        resolved_scope,
        load_policy,
        gang_scheduled_task=False,
        temp_dir_root=None,
    ):
        self._chkptr = CurrentCheckpointer(
            flow=flow,
            task_identifier=default_task_identifier,
            resolved_scope=resolved_scope,
            logger=self._logger,
            gang_scheduled_task=gang_scheduled_task,
            temp_dir_root=temp_dir_root,
        )
        return self._chkptr._setup_task_first_load(load_policy, flow)

    def _setup_current(
        self,
    ):
        from metaflow import current

        current._update_env(
            {
                "checkpoint": self._chkptr,
            }
        )



STEP_DECORATORS_DESC = [
    (
        "checkpoint",
        ".machine_learning_utilities.checkpoints.decorator.CheckpointDecorator",
    ),
    (
        "model",
        ".machine_learning_utilities.modeling_utils.decorator.ModelDecorator",
    ),
    (
        "huggingface_hub",
        ".machine_learning_utilities.hf_hub.decorator.HuggingfaceHubDecorator",
    ),
]

###
# CONFIGURE: Similar to datatools, you can make visible under metaflow.plugins.* other
#            submodules not referenced in this file
###
__mf_promote_submodules__ = []



import os
from metaflow.cards import VegaChart, MetaflowCardComponent, Artifact, Table
from typing import Optional, List, Union
import math
from metaflow.plugins.cards.card_modules.components import (
    with_default_component_id,
    TaskToDict,
    ArtifactsComponent,
    UserComponent,
    TableComponent,
    SectionComponent,
    render_safely,
)
import datetime
from metaflow.metaflow_current import current
import json
from functools import wraps
from collections import defaultdict
from threading import Thread, Event
import time


def update_spec_data(spec, data):
    spec["data"]["values"] = data
    return spec


def update_data_object(data_object, data):
    data_object["values"].append(data)
    return data_object


def line_chart_spec(
    title=None,
    x_name="u",
    y_name="v",
    xtitle=None,
    ytitle=None,
    width=600,
    height=400,
    with_params=True,
    x_axis_temporal=False,
):
    parameters = [
        {
            "name": "interpolate",
            "value": "linear",
            "bind": {
                "input": "select",
                "options": [
                    "basis",
                    "cardinal",
                    "catmull-rom",
                    "linear",
                    "monotone",
                    "natural",
                    "step",
                    "step-after",
                    "step-before",
                ],
            },
        },
        {
            "name": "tension",
            "value": 0,
            "bind": {"input": "range", "min": 0, "max": 1, "step": 0.05},
        },
        {
            "name": "strokeWidth",
            "value": 2,
            "bind": {"input": "range", "min": 0, "max": 10, "step": 0.5},
        },
        {
            "name": "strokeCap",
            "value": "butt",
            "bind": {"input": "select", "options": ["butt", "round", "square"]},
        },
        {
            "name": "strokeDash",
            "value": [1, 0],
            "bind": {
                "input": "select",
                "options": [[1, 0], [8, 8], [8, 4], [4, 4], [4, 2], [2, 1], [1, 1]],
            },
        },
    ]
    parameter_marks = {
        "interpolate": {"expr": "interpolate"},
        "tension": {"expr": "tension"},
        "strokeWidth": {"expr": "strokeWidth"},
        "strokeDash": {"expr": "strokeDash"},
        "strokeCap": {"expr": "strokeCap"},
    }
    spec = {
        "title": title if title else "Line Chart",
        "$schema": "https://vega.github.io/schema/vega-lite/v5.json",
        # "width": width,
        # "height": height,
        "params": parameters if with_params else [],
        "data": {"name": "values", "values": []},
        "mark": {
            "type": "line",
            "tooltip": True,
            **(parameter_marks if with_params else {}),
        },
        "selection": {"grid": {"type": "interval", "bind": "scales"}},
        "encoding": {
            "x": {
                "field": x_name,
                "title": xtitle if xtitle else x_name,
                **({"timeUnit": "seconds"} if x_axis_temporal else {}),
                **({"type": "quantitative"} if not x_axis_temporal else {}),
            },
            "y": {
                "field": y_name,
                "type": "quantitative",
                "title": ytitle if ytitle else y_name,
            },
        },
        **({"width": width, "height": height} if width and height else {}),
    }
    data = {"values": []}
    return spec, data


class LineChart(UserComponent):
    REALTIME_UPDATABLE = True

    def __init__(
        self,
        title,
        xtitle,
        ytitle,
        x_name,
        y_name,
        with_params=False,
        x_axis_temporal=False,
        width=None,
        height=None,
    ):
        super().__init__()

        self.spec, _ = line_chart_spec(
            title=title,
            xtitle=xtitle,
            ytitle=ytitle,
            x_name=x_name,
            y_name=y_name,
            with_params=with_params,
            x_axis_temporal=x_axis_temporal,
            width=width,
            height=height,
        )

    def update(self, data):  # Can take a diff
        self.spec = update_spec_data(self.spec, data)

    @with_default_component_id
    def render(self):
        vega_chart = VegaChart(self.spec, show_controls=True)
        vega_chart.component_id = self.component_id
        return vega_chart.render()


class ArtifactTable(Artifact):
    def __init__(self, data_dict):
        self._data = data_dict
        self._task_to_dict = TaskToDict(only_repr=True)

    @with_default_component_id
    @render_safely
    def render(self):
        _art_list = []
        for k, v in self._data.items():
            _art = self._task_to_dict.infer_object(v)
            _art["name"] = k
            _art_list.append(_art)

        af_component = ArtifactsComponent(data=_art_list)
        af_component.component_id = self.component_id
        return af_component.render()


class UpadateableTable(UserComponent):
    """

    Parameters
    ----------
    data : List[List[str or MetaflowCardComponent]], optional
        List (rows) of lists (columns). Each item can be a string or a `MetaflowCardComponent`.
    headers : List[str], optional
        Optional header row for the table.
    """

    REALTIME_UPDATABLE = True

    def update(self, row: List[Union[str, MetaflowCardComponent]]):
        self._data.append(row)

    def __init__(
        self,
        data: Optional[List[List[Union[str, MetaflowCardComponent]]]] = None,
        headers: Optional[List[str]] = None,
        disable_updates: bool = False,
    ):
        data = data or [[]]
        headers = headers or []
        header_bool, data_bool = TableComponent.validate(headers, data)
        self._headers = []
        self._data = [[]]
        if header_bool:
            self._headers = headers
        if data_bool:
            self._data = data

        if disable_updates:
            self.REALTIME_UPDATABLE = False

    def _render_subcomponents(self):
        for row in self._data:
            for col in row:
                if isinstance(col, VegaChart):
                    col._chart_inside_table = True

        return [
            SectionComponent.render_subcomponents(
                row,
                additional_allowed_types=[
                    str,
                    bool,
                    int,
                    float,
                    dict,
                    list,
                    tuple,
                    type(None),
                ],
                allow_unknowns=True,
            )
            for row in self._data
        ]

    @with_default_component_id
    @render_safely
    def render(self):
        table_component = TableComponent(
            headers=self._headers, data=self._render_subcomponents()
        )
        table_component.component_id = self.component_id
        return table_component.render()



from hashlib import sha256
import json
from typing import TYPE_CHECKING
from metaflow.exception import MetaflowException
from .constants import MAX_HASH_LEN


if TYPE_CHECKING:
    import metaflow


class AttemptEnum:
    LATEST = "latest"
    PREVIOUS = "previous"
    FIRST = "first"

    @classmethod
    def values(cls):
        return [cls.LATEST, cls.PREVIOUS, cls.FIRST]


class IdentityException(MetaflowException):
    pass


class UnhashableValueException(MetaflowException):
    def __init__(self, type):
        self._type = type
        super().__init__(f"Value of type %s is not safely hashable." % type)


class FlowNotRunningException(MetaflowException):
    pass


def pathspec_hash(pathspec: str):
    return safe_hash(pathspec)[:8]


def safe_hash(value):
    """Safely hash arbitrary values and a created a sha for that object, including those that are not natively hashable."""
    try:
        # Try to hash the value directly, works for immutable objects
        return sha256(hash(value)).hexdigest()
    except TypeError:
        # If direct hashing fails, serialize the object and hash the serialized string
        # This is slower but will work for un-hashable objects like lists and dicts
        try:
            serialized = json.dumps(value, sort_keys=True).encode()
            return sha256(serialized).hexdigest()
        except TypeError as e:
            # Handle the case where the object cannot be serialized
            raise UnhashableValueException(str(type(value)))


def _resolve_hash_of_current_foreach_stack(stack):

    stack_of_values = [x[2] for x in stack]

    # We add a value to the stack which is a deterministic value that is random enough that clashes become unlikely.
    # Otherwise if the foreach value can contain a value that is same as the stepname, then the hash of the foreach
    # value and the stepname will be the same causing a potential state clash
    stack_of_values.append(safe_hash("foreach-task-entropy"))
    # create a hash of all the inputs in the stack.
    hashed_inputs = "".join([safe_hash(x) for x in stack_of_values])

    foreach_task_identifier = sha256((hashed_inputs).encode()).hexdigest()

    return foreach_task_identifier


def _resolve_hash_for_particular_parallel_index(
    run: "metaflow.FlowSpec",
    index,
):
    # It is assumed here that the function is only called for a run
    # with a parallel task.
    stack = run.foreach_stack()
    # to derive the task identifier of the gang scheduled parallel jobs
    # we need to ensure that all tasks within the gang get the same task identifer.
    # This value is based on the index of the task provided to this function. (for most default cases it will be 0)
    # We achieve this be poping the index of the foreach stack and replacing it with the index of the task.
    stack_of_values = [x[2] for x in stack]
    stack_of_values.pop()
    # When we replace the index of the foreach stack, we can also potentially hit a situation where the task
    # is converted from the parallel task to a regular foreach task. At this point we need to ensure that the
    # task identifier is not the same as the parallel task. To ensure this, we add a value to the stack which
    # is a deterministic value that is random enough that clashes become unlikely.
    stack_of_values.extend([index, safe_hash("parallel-task-entropy")])

    # create a hash of all the inputs in the stack.
    hashed_inputs = "".join([safe_hash(x) for x in stack_of_values])

    parallel_task_identifier = sha256((hashed_inputs).encode()).hexdigest()

    return parallel_task_identifier


class TaskIdentifier:
    """
    Any change to this class's core logic can create SEVERE backwards compatibility issues
    since this class helps derive the task identifier for the checkpoints.

    IDEALLY, the identifier construction logic of this file should be kept as is.
    """

    @classmethod
    def for_parallel_task_index(cls, run: "metaflow.FlowSpec", index: int):
        """
        This class is meant to mint a task-identifier for a parallel task based on the
        index of the task in the gang.
        """
        base_task_identifier = _resolve_hash_for_particular_parallel_index(run, index)
        return base_task_identifier[:MAX_HASH_LEN]

    @classmethod
    def from_flowspec(cls, run: "metaflow.FlowSpec"):
        from metaflow import current

        if not current.is_running_flow:
            raise FlowNotRunningException

        base_task_identifier = None
        # Only resolving identifier for foreach value first.
        # If it is not present, then we will resolve the identifier
        # based on the stepname.
        if run.index is not None:
            stack = run.foreach_stack()
            base_task_identifier = _resolve_hash_of_current_foreach_stack(stack)
        else:
            # If it is not a foreach task, the the identifier can be
            # set a static value (like hash of the stepname) since it will
            # make list prefixing a lot simpler and the logic of understanding
            # base identity a lot simpler. Otherwise we will have
            # to distinguish between how we resolve the identifier for
            # foreach and non-foreach tasks in different usecases.
            step_name = current.step_name
            base_task_identifier = safe_hash(step_name)

        return base_task_identifier[:MAX_HASH_LEN]

    def from_task(cls, task: "metaflow.Task", use_origin=True):

        return safe_hash(task.parent.id)[:MAX_HASH_LEN]



"""
This file contains methods injected into the `DataStoreStorage` class
"""
from metaflow.plugins.datastores.s3_storage import S3, ARTIFACT_LOCALROOT, CloseAfterUse
from metaflow.plugins.datastores.gs_storage import (
    handle_executor_exceptions,
    as_completed,
)
from metaflow.plugins.datatools.s3 import S3PutObject
from metaflow.plugins.datastores.local_storage import LocalStorage
import shutil
import os


def load_files_s3(self, keys):
    return _s3_load_bytes(self, keys)


def load_files_gcp_or_azure_or_local(self, keys):
    return self.load_bytes(keys)


def _s3_load_bytes(self, paths):
    if len(paths) == 0:
        return CloseAfterUse(iter([]))

    s3 = S3(
        s3root=self.datastore_root,
        external_client=self.s3_client,
    )

    def iter_results():
        # We similarly do things in parallel for many files. This is again
        # a hack.
        if len(paths) > 10:
            results = s3.get_many(paths, return_missing=True, return_info=True)
            for r in results:
                if r.exists:
                    yield r.key, r.path, r.metadata
                else:
                    yield r.key, None, None
        else:
            for p in paths:
                r = s3.get(p, return_missing=True, return_info=True)
                if r.exists:
                    yield r.key, r.path, r.metadata
                else:
                    yield r.key, None, None

    return CloseAfterUse(iter_results(), closer=s3)


def save_files_s3(self, key_path_tuples, overwrite=False):
    with S3(
        s3root=self.datastore_root,
        external_client=self.s3_client,
    ) as s3:
        s3.put_files(
            [
                S3PutObject(
                    key=key,
                    path=path,
                )
                for key, path in key_path_tuples
            ],
            overwrite=overwrite,
        )


def save_file_s3(self, key, path, overwrite=False):
    with S3(
        s3root=self.datastore_root,
        external_client=self.s3_client,
    ) as s3:
        s3.put_files(
            [
                S3PutObject(
                    key=key,
                    path=path,
                )
            ],
            overwrite=overwrite,
        )


@handle_executor_exceptions
def save_files_gcp_or_azure(self, key_path_tuples, overwrite=False):
    futures = []
    for key, path in key_path_tuples:
        futures.append(
            self._executor.submit(
                self.root_client.save_bytes_single,
                (key, path, {"save_file_single": True}),
                overwrite=overwrite,
            )
        )
    for future in as_completed(futures):
        future.result()


@handle_executor_exceptions
def save_file_gcp_or_azure(self, key, path, overwrite=False):
    futures = []
    futures.append(
        self._executor.submit(
            self.root_client.save_bytes_single,
            (key, path, {"save_file_single": True}),
            overwrite=overwrite,
        )
    )
    for future in as_completed(futures):
        future.result()


def save_file_local(self, key, path, overwrite=False):

    full_path = self.full_uri(key)
    if not overwrite and os.path.exists(full_path):
        return
    LocalStorage._makedirs(os.path.dirname(full_path))
    shutil.copy(path, full_path)


def save_files_local(self, key_path_tuples, overwrite=False):
    for key, path in key_path_tuples:
        save_file_local(self, key, path, overwrite=overwrite)


STORAGE_INJECTIONS_SINGLE_FILE_SAVE = {
    "s3": save_file_s3,
    "gs": save_file_gcp_or_azure,
    "azure": save_file_gcp_or_azure,
    "local": save_file_local,
}

STORAGE_INJECTIONS_MULTIPLE_FILE_SAVE = {
    "s3": save_files_s3,
    "gs": save_files_gcp_or_azure,
    "azure": save_files_gcp_or_azure,
    "local": save_files_local,
}

STORAGE_INJECTIONS_LOAD_FILES = {
    "s3": load_files_s3,
    "gs": load_files_gcp_or_azure_or_local,
    "azure": load_files_gcp_or_azure_or_local,
    "local": load_files_gcp_or_azure_or_local,
}



import os
from typing import Callable, Generator, Iterator, List, Optional, Union

from metaflow.datastore.datastore_storage import DataStoreStorage
from .constants import DEFAULT_NAME, CHECKPOINTS_STORAGE_PREFIX, DEFAULT_STORAGE_FORMAT
from ..exceptions import KeyNotCompatibleWithObjectException
from ..utils.identity_utils import pathspec_hash
from ..utils.general import replace_start_and_end_slash
from ..datastore.core import (
    allow_safe,
    DatastoreInterface,
    ObjectStorage,
    STORAGE_FORMATS,
)
from ..datastore.exceptions import (
    DatastoreReadInitException,
    DatastoreWriteInitException,
    DatastoreNotReadyException,
)
from ..datastructures import CheckpointArtifact
from ..datastore.utils import safe_serialize
import json
import re
from datetime import datetime
from collections import namedtuple

ARTIFACT_STORE_NAME = "artifacts"

METADATA_STORE_NAME = "metadata"

ARTIFACT_METADATA_STORE_NAME = "artifact_metadata"


CheckpointsPathComponents = namedtuple(
    "CheckpointsPathComponents",
    [
        "flow_name",
        "step_name",
        "scope",
        "task_identifier",
        "pathspec_hash",
        "attempt",
        "name",
        "version_id",
        "is_metadata",
        "key_name",  # This is the final part of the path components and is generally of the form `a.b.c.d.metadata` or `a.b.c.d`
        "root_prefix",  # This is the root prefix of the path like `mf.checkpoints`
    ],
)

# Every checkpoint key will be prefixed with a `/checkpoint/` and will have a structure of the form:
# <root_prefix>/checkpoints/artifacts/<flow_name>/<step_name>/<scope>/<task_identifier>/<pathspec_hash>.<attempt>.<name>.<version_id>
# The root_prefix can be set at the decorator level and the /checkpoints/ is a fixed prefix that also helps distinguish between the
# different types of objects such as `checkpoints`/`models` etc.
CHECKPOINT_KEY_PATTERN = re.compile(
    r"^(?P<root_prefix>.*)/checkpoints/artifacts/(?P<flow_name>.*?)/(?P<step_name>.*?)/(?P<scope>.*?)/(?P<task_identifier>.*?)/(?P<key_name>.*?)$"
)

CHECKPOINT_ARTIFACT_MD_KEY_PATTERN = re.compile(
    r"^(?P<root_prefix>.*)/checkpoints/artifact_metadata/(?P<flow_name>.*?)/(?P<step_name>.*?)/(?P<scope>.*?)/(?P<task_identifier>.*?)/(?P<key_name>.*?)$"
)


CHECKPOINT_METADATA_KEY_PATTERN = re.compile(
    r"^(?P<root_prefix>.*)/checkpoints/metadata/(?P<flow_name>.*?)/(?P<runid>.*?)/(?P<step_name>.*?)/(?P<taskid>.*?)/(?P<key_name>.*?)$"
)


def __decompose_checkpoint_artifact_metadata_key(key):
    # Match the string with the pattern
    match = re.match(CHECKPOINT_ARTIFACT_MD_KEY_PATTERN, key)
    # Extract the values into a dictionary if there's a match
    if match:
        values = match.groupdict()
        return values

    raise KeyNotCompatibleWithObjectException(key, "checkpoint_artifact_metadata")


def __decompose_checkpoint_metadata_key(key):
    # Match the string with the pattern
    match = re.match(CHECKPOINT_METADATA_KEY_PATTERN, key)
    # Extract the values into a dictionary if there's a match
    if match:
        values = match.groupdict()
        return values

    raise KeyNotCompatibleWithObjectException(key, "checkpoint_metadata")


def __decompose_checkpoint_key(key):
    # Match the string with the pattern
    match = re.match(CHECKPOINT_KEY_PATTERN, key)
    # Extract the values into a dictionary if there's a match
    if match:
        values = match.groupdict()
        return values

    raise KeyNotCompatibleWithObjectException(key, "checkpoint")


def _decompose_artifact_key(key, data_object):
    root_prefix, flowname, stepname, scope, task_identifier, chckpt_id = (
        data_object["root_prefix"],
        data_object["flow_name"],
        data_object["step_name"],
        data_object["scope"],
        data_object["task_identifier"],
        data_object["key_name"],
    )

    if len(chckpt_id.split(".")) < 4:
        raise KeyNotCompatibleWithObjectException(
            key,
            "checkpoint",
            "key_name is not in the correct format. Expected 4 parts, got %s"
            % len(chckpt_id.split(".")),
        )

    is_metadata = chckpt_id.split(".")[-1] == "metadata"
    key_name = chckpt_id
    if is_metadata:
        key_name = ".".join(chckpt_id.split(".")[:-1])

    pathspec_hash, attempt, name, version_id = chckpt_id.split(".")[:4]
    return CheckpointsPathComponents(
        flow_name=flowname,
        step_name=stepname,
        scope=scope,
        task_identifier=task_identifier,
        pathspec_hash=pathspec_hash,
        attempt=attempt,
        name=name,
        version_id=version_id,
        is_metadata=is_metadata,
        key_name=key_name,
        root_prefix=root_prefix,
    )


def decompose_key_artifact_metadata_store(
    key,
) -> CheckpointsPathComponents:
    data_object = __decompose_checkpoint_artifact_metadata_key(key)
    return _decompose_artifact_key(key, data_object)


def decompose_key_artifact_store(
    key,
) -> CheckpointsPathComponents:
    """
    Convert Key into Path Components.
    PATH COMPONENTS: mf.checkpoints/artifacts/<flow_name>/<step_name>/<scope>/<task_identifier>/<pathspec_hash>.<attempt>.<name>.<version_id>

    """
    data_object = __decompose_checkpoint_key(key)
    return _decompose_artifact_key(key, data_object)


def decompose_key_metadata_store(
    key,
) -> CheckpointsPathComponents:
    """
    Convert Key into Path Components.
    PATH COMPONENTS: mf.checkpoints/artifacts/<flow_name>/<step_name>/<scope>/<task_identifier>/<pathspec_hash>.<attempt>.<name>.<version_id>

    """
    _data = __decompose_checkpoint_metadata_key(key)
    root_prefix, flowname, runid, stepname, taskid, chckpt_id = (
        _data["root_prefix"],
        _data["flow_name"],
        _data["runid"],
        _data["step_name"],
        _data["taskid"],
        _data["key_name"],
    )
    if len(chckpt_id.split(".")) < 3:
        raise KeyNotCompatibleWithObjectException(
            key,
            "checkpoint_metadata",
            "key_name is not in the correct format. Expected 3 parts, got %s"
            % len(chckpt_id.split(".")),
        )

    is_metadata = chckpt_id.split(".")[-1] == "metadata"
    key_name = chckpt_id
    if is_metadata:
        key_name = ".".join(chckpt_id.split(".")[:-1])

    attempt, name, version_id = chckpt_id.split(".")[:3]
    return CheckpointsPathComponents(
        flow_name=flowname,
        step_name=stepname,
        scope=None,
        task_identifier=None,
        pathspec_hash=pathspec_hash("/".join([flowname, runid, stepname, taskid])),
        attempt=attempt,
        name=name,
        version_id=version_id,
        is_metadata=is_metadata,
        key_name=key_name,
        root_prefix=root_prefix,
    )


class CheckpointDatastore(DatastoreInterface):

    """
    Consisits of 3 main components:
    - `artifact_store`: This is where the checkpoint artifacts are stored.
        - This key to the checkpoint in this store becomes the "key for the checkpoint"
    - `metadata_store`: This is where the metadata of the checkpoint is stored it based on the currently executing task (path structure resembles that of a metaflow pathspec).
        - This store helps retrieve information about all the checkpoints stored for a task during the execution.
    - `artifact_metadatastore`: This is similar to the metadata store but holds a pathstructure similar to the artifact store.
        - this store helps reverse lookup the Checkpoint metadata object from the checkpoint key.
    """

    ROOT_PREFIX = CHECKPOINTS_STORAGE_PREFIX

    artifact_store: ObjectStorage = None

    metadata_store: ObjectStorage = None

    artifact_metadatastore: ObjectStorage = None

    _NAME_ENTROPY = None

    pathspec = None

    STORAGE_FORMATS = [
        STORAGE_FORMATS.TAR,
        STORAGE_FORMATS.FILES
        # can add more as needed.
    ]

    @property
    def metadata_ready(self):
        return self.metadata_store is not None

    @property
    def artifact_ready(self):
        return self.artifact_store is not None

    def set_root_prefix(self, root_prefix):
        self.ROOT_PREFIX = root_prefix
        if self.metadata_store is not None:
            self.metadata_store.set_full_prefix(root_prefix)
        if self.artifact_store is not None:
            self.artifact_store.set_full_prefix(root_prefix)
        if self.artifact_metadatastore is not None:
            self.artifact_metadatastore.set_full_prefix(root_prefix)

    @classmethod
    def init_read_store(
        cls,
        storage_backend: DataStoreStorage,
        pathspec=None,
        checkpoint_key=None,
    ):
        """
        This will initialize the datastore for reading.

        - If there is only the pathspec that's provided then it can mean the user is doing a list operations
        - if only the checkpoint_key is provided then it can mean the user is trying to load a specific checkpoint
        """
        datastore = cls()
        if pathspec is not None:
            datastore.metadata_store = ObjectStorage(
                storage_backend,
                root_prefix=cls.ROOT_PREFIX,
                path_components=[
                    "checkpoints",
                    METADATA_STORE_NAME,
                    *pathspec.split("/"),
                ],
            )
            datastore.pathspec = pathspec

            datastore._NAME_ENTROPY = pathspec_hash(pathspec)
        elif checkpoint_key is not None:
            _key_decomp = cls.decompose_key(checkpoint_key)
            _path_components = [
                _key_decomp.flow_name,
                _key_decomp.step_name,
                _key_decomp.scope,
                _key_decomp.task_identifier,
            ]
            datastore.artifact_store = ObjectStorage(
                storage_backend,
                root_prefix=cls.ROOT_PREFIX,
                path_components=["checkpoints", ARTIFACT_STORE_NAME] + _path_components,
            )
            datastore.artifact_metadatastore = ObjectStorage(
                storage_backend,
                root_prefix=cls.ROOT_PREFIX,
                path_components=["checkpoints", ARTIFACT_METADATA_STORE_NAME]
                + _path_components,
            )
            datastore._NAME_ENTROPY = _key_decomp.pathspec_hash
            datastore.set_root_prefix(_key_decomp.root_prefix)
        else:
            raise DatastoreReadInitException(
                "pathspec or checkpoint_key must be provided"
            )

        return datastore

    def create_key_name(
        self,
        *args,
    ):
        return ".".join([str(a) for a in args])

    @classmethod
    def init_write_store(
        cls,
        storage_backend: DataStoreStorage,
        pathspec,
        scope,
        task_identifier,
    ):
        if any([pathspec is None, scope is None, task_identifier is None]):
            raise DatastoreWriteInitException(
                "pathspec, scope, task_identifier must be provided"
            )
        datastore = cls()
        flow_name, runid, step_name, taskid = pathspec.split("/")

        datastore.metadata_store = ObjectStorage(
            storage_backend,
            root_prefix=cls.ROOT_PREFIX,
            path_components=[
                "checkpoints",
                METADATA_STORE_NAME,
                flow_name,
                runid,
                step_name,
                taskid,
            ],
        )
        datastore.pathspec = pathspec

        datastore._NAME_ENTROPY = pathspec_hash(pathspec)
        path_components = [
            flow_name,
            step_name,
            scope,
            task_identifier,
        ]
        datastore.artifact_store = ObjectStorage(
            storage_backend,
            root_prefix=cls.ROOT_PREFIX,
            path_components=["checkpoints", ARTIFACT_STORE_NAME] + path_components,
        )

        datastore.artifact_metadatastore = ObjectStorage(
            storage_backend,
            root_prefix=cls.ROOT_PREFIX,
            path_components=["checkpoints", ARTIFACT_METADATA_STORE_NAME]
            + path_components,
        )
        return datastore

    def save(
        self,
        local_path: str,
        attempt,
        version_id,
        name=DEFAULT_NAME,
        metadata={},
        set_latest=True,
        storage_format=DEFAULT_STORAGE_FORMAT,
    ) -> CheckpointArtifact:

        if not (self.artifact_ready and self.metadata_ready):
            raise DatastoreNotReadyException(
                "Checkpoints Datastore is not ready for write operations"
            )

        _key = self.create_key_name(
            self._NAME_ENTROPY,
            attempt,
            name,
            version_id,
            # AT Write TIME pathspec hash is AlWAYS RESOLVABLE
            # because `metadata_store` is ALWAYS SET
        )

        _storage_func = None
        if storage_format == STORAGE_FORMATS.FILES:
            _storage_func = self.artifact_store._save_objects
        elif storage_format == STORAGE_FORMATS.TAR:
            _storage_func = self.artifact_store._save_tarball
        else:
            raise ValueError(
                "Invalid storage format. Expected one of %s got %s"
                % (self.STORAGE_FORMATS, storage_format)
            )

        (full_checkpoint_url, key_path, file_size,) = _storage_func(
            _key,
            local_path,
        )

        _metadata = dict(
            size=file_size,
            pathspec=self.pathspec,
            pathspec_hash=self._NAME_ENTROPY,
            attempt=attempt,
            key=key_path,
            type=CheckpointArtifact.TYPE,
            url=full_checkpoint_url,
            name=name,
            created_on=datetime.now().isoformat(),
            metadata=safe_serialize(metadata),
            storage_format=storage_format,
            creation_context="task",
            version_id=version_id,
        )

        _art_key = self.create_key_name(
            self._NAME_ENTROPY,
            attempt,
            name,
            version_id,
            "metadata",
        )
        _md_key = self.create_key_name(
            attempt,
            name,
            version_id,
            "metadata",
        )
        for _key, store in zip(
            [_art_key, _md_key],
            [
                self.artifact_metadatastore,
                self.metadata_store,
            ],
        ):
            store._save_metadata(_key, _metadata)
            if set_latest:
                store._save_metadata("latest", _metadata)
        return CheckpointArtifact.from_dict(_metadata)

    @allow_safe
    def latest(self, current_task=True) -> CheckpointArtifact:
        if current_task:
            if not self.metadata_ready:
                raise DatastoreNotReadyException(
                    "Checkpoint store is not ready to read the latest checkpoint in current task"
                )
            _md = self.metadata_store._load_metadata("latest")
        else:
            if not self.artifact_ready:
                raise DatastoreNotReadyException(
                    "Checkpoint store is not ready to read the latest checkpoint"
                )
            _md = self.artifact_metadatastore._load_metadata("latest")

        return CheckpointArtifact.from_dict(_md)

    def load(
        self,
        local_path,
        version_id,
        attempt,
        name,
        storage_format=DEFAULT_STORAGE_FORMAT,
    ):
        if not self.artifact_ready:
            raise ValueError("Datastore is not ready to load")
        _key = self.create_key_name(
            self._NAME_ENTROPY,
            attempt,
            name,
            version_id,
        )
        _load_func = None
        if storage_format == "tar":
            return self.artifact_store._load_tarball(_key, local_path)
        elif storage_format == "files":
            return self.artifact_store._load_objects(_key, local_path)
        raise ValueError(
            "Invalid storage format. Expected one of %s got %s"
            % (self.STORAGE_FORMATS, storage_format)
        )

    def load_metadata(
        self,
        attempt,
        version_id,
        name=DEFAULT_NAME,
    ) -> dict:
        if self.metadata_ready:
            return self.metadata_store._load_metadata(
                self.create_key_name(
                    attempt,
                    name,
                    version_id,
                    "metadata",
                )
            )
        elif self.artifact_ready:
            return self.artifact_metadatastore._load_metadata(
                self.create_key_name(
                    self._NAME_ENTROPY,
                    attempt,
                    name,
                    version_id,
                    "metadata",
                )
            )
        raise DatastoreNotReadyException("Datastore is not ready to load metadata")

    def list(
        self,
        name=None,
        attempt=None,
        within_task=True,
    ):

        if not within_task:
            if not self.artifact_ready:
                raise DatastoreNotReadyException(
                    "Checkpoint datastore is not ready to list all checkpoints"
                )
            return _recover_checkpoints(
                self.artifact_metadatastore,
                key_decomposer=decompose_key_artifact_metadata_store,
                name=name,
                attempt=attempt,
            )

        if not self.metadata_ready:
            raise DatastoreNotReadyException(
                "Checkpoint datastore is not ready to list checkpoints within the task"
            )
        return _recover_checkpoints(
            self.metadata_store,
            key_decomposer=decompose_key_metadata_store,
            name=name,
            attempt=attempt,
        )

    @classmethod
    def decompose_key(cls, key) -> CheckpointsPathComponents:
        return decompose_key_artifact_store(key)


def _recover_checkpoints(
    datastore: ObjectStorage,
    key_decomposer: Callable[[str], CheckpointsPathComponents],
    name: Optional[str] = None,
    attempt: Optional[int] = None,
) -> Iterator[CheckpointArtifact]:
    def _validate_name(info: CheckpointsPathComponents):
        if name is not None and info.name != name:
            return False
        return True

    def _filter_based_on_attempts(info: CheckpointsPathComponents):
        if attempt is None:
            return True
        if type(attempt) == int:
            return str(info.attempt) == str(attempt)
        return True

    # `datastore.list_paths` will have very different outputs based on the type of datastore.
    # - for CheckpointMetadataStore it will list ALL checkpoints within the task.
    # - for CheckpointArtifactMetadataStore it will list ALL checkpoints within scope/the task-identifier.
    #   - Meaning that for retrieving checkpoints during retries/re-executions can be a lot faster.

    # If we want we can even list within only-scope by using the
    # CheckpointArtifactStore it will list ALL checkpoints within the "scope"
    # (this can be astoundingly large if things are running inside foreaches)

    for path_tup in datastore.list_paths([""]):
        try:
            obj_info = key_decomposer(
                path_tup.key,
            )
        except KeyNotCompatibleWithObjectException as e:
            # this means that we hit an object that might be a reference but not something we are looking for
            continue
        if not _filter_based_on_attempts(obj_info):
            continue
        if not _validate_name(obj_info):
            continue
        metadata = datastore._load_metadata(
            datastore.create_key_name(obj_info.key_name, "metadata")
        )
        yield CheckpointArtifact.from_dict(metadata)



from setuptools import setup, find_namespace_packages

version = "0.1.5"


def get_long_description() -> str:
    with open("README.md") as fh:
        return fh.read()


setup(
    name="metaflow-checkpoint",
    version=version,
    description="An EXPERIMENTAL checkpoint decorator for Metaflow",
    author="Valay Dave",
    author_email="help@outerbounds.com",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    packages=find_namespace_packages(include=["metaflow_extensions.*"]),
    py_modules=[
        "metaflow_extensions",
    ],
    install_requires=[],
)



__mf_extensions__ = "obcheckpoint"

from ..plugins.machine_learning_utilities.checkpoints.final_api import Checkpoint
from ..plugins.machine_learning_utilities.datastructures import load_model

try:
    import pkg_resources

    __version__ = pkg_resources.get_distribution("metaflow-checkpoint").version
except:
    # this happens on remote environments since the job package
    # does not have a version
    __version__ = None



from typing import List, Dict, Union, Tuple, Optional, TYPE_CHECKING
from .core import (
    Checkpointer,
    WriteResolver,
    ReadResolver,
)
from ..datastructures import CheckpointArtifact
from .constants import CHECKPOINT_UID_ENV_VAR_NAME, DEFAULT_NAME
import os

if TYPE_CHECKING:
    import metaflow
    from .final_api import Checkpoint


def _instantiate_checkpoint(
    checkpoint_object: "Checkpoint",
    flow: Optional["metaflow.FlowSpec"] = None,  # User Facing
    task_identifier: Optional[str] = None,  # Advanced options
    scope: Optional[str] = None,  # Advanced options
    gang_scheduled_task: bool = False,
):
    if flow is not None:
        datastore, resolver_info = WriteResolver.from_run(
            flow,
            scope=scope,
            task_identifier=task_identifier,
            gang_scheduled_task=gang_scheduled_task,
        )
    elif CHECKPOINT_UID_ENV_VAR_NAME in os.environ:
        datastore, resolver_info = WriteResolver.from_environment()
    else:
        raise ValueError(
            "Creating a `Checkpoint` object requires either a `FlowSpec` instance or a `METAFLOW_CHECKPOINT_UID` environment variable."
        )

    # Technically at this point `resolver_info` contains all the information to create
    # any new checkpointer capable of writing.
    if any(
        a is None for a in resolver_info
    ):  # Ideally we should NOT be hitting this line at ALL!
        raise ValueError(
            "Missing enough information to instantiate a Checkpoint object."
        )
    chkpt = Checkpointer(
        datastore=datastore,
        attempt=resolver_info.attempt,
    )
    chkpt._checkpoint_uid = WriteResolver.construct_checkpoint_id(resolver_info)
    checkpoint_object._set_checkpointer(chkpt)
    return checkpoint_object


# Technically this method should ensure that the Checkpointer also takes the
# attempts as a part of the `load` method.
# READ METHOD
def load_checkpoint(  # Return load status
    checkpoint: Union[CheckpointArtifact, dict, str],  # User Facing
    local_path: str,  # User Facing
):
    # It's assumed that everything is already correctly passsed
    if isinstance(checkpoint, CheckpointArtifact) or isinstance(checkpoint, dict):
        _load_checkpoint_from_reference(checkpoint, local_path)
    elif isinstance(checkpoint, str):
        _load_checkpoint_from_key(checkpoint, local_path)
    else:
        raise ValueError(
            "Invalid checkpoint object. Expected CheckpointArtifact, dict or str. Got: %s"
            % type(checkpoint)
        )


def _load_checkpoint_from_reference(
    checkpoint: Union[CheckpointArtifact, dict],  # User Facing
    local_path: str,  # User Facing
):
    _chckpt = CheckpointArtifact.hydrate(checkpoint)
    # this will already set the version on the checkpointer.
    _checkpointer = Checkpointer._from_checkpoint(_chckpt)
    _name = _chckpt.name
    version = _checkpointer.current_version
    _checkpointer._load_checkpoint(
        local_path,
        version_id=version,
        name=_name,
        storage_format=_chckpt.storage_format,
    )


def _load_checkpoint_from_key(
    checkpoint_key: str,
    local_path: str,
):
    _object = CheckpointArtifact._load_metadata_from_key(checkpoint_key, None)
    _checkpointer = Checkpointer._from_checkpoint(_object)

    _checkpointer._load_checkpoint(
        local_path,
        version_id=_object.version_id,
        name=_object.name,
        storage_format=_object.storage_format,
    )


def _coalesce(*args):
    for arg in args:
        if arg is not None:
            return arg
    return None
